@@ -11,11 +11,20 @@
 #    License for the specific language governing permissions and limitations
 #    under the License.
 
+import argparse
 import os
 import sys
 import traceback
 import yaml
 
+# Only permit the template alias versions
+# The current template version should be the last element
+valid_heat_template_versions = [
+  'ocata',
+  'pike',
+  'queens',
+]
+current_heat_template_version = valid_heat_template_versions[-1]
 
 required_params = ['EndpointMap', 'ServiceNetMap', 'DefaultPasswords',
                    'RoleName', 'RoleParameters', 'ServiceData']
@@ -33,16 +42,21 @@ envs_containing_endpoint_map = ['tls-endpoints-public-dns.yaml',
 ENDPOINT_MAP_FILE = 'endpoint_map.yaml'
 OPTIONAL_SECTIONS = ['workflow_tasks', 'cellv2_discovery']
 REQUIRED_DOCKER_SECTIONS = ['service_name', 'docker_config', 'puppet_config',
-                            'config_settings', 'step_config']
+                            'config_settings']
 OPTIONAL_DOCKER_SECTIONS = ['docker_puppet_tasks', 'upgrade_tasks',
-                            'post_upgrade_tasks', 'update_tasks',
-                            'service_config_settings',
-                            'host_prep_tasks', 'metadata_settings',
-                            'kolla_config', 'logging_source',
-                            'logging_groups', 'docker_config_scripts']
+                            'fast_forward_upgrade_tasks',
+                            'post_upgrade_tasks',  'update_tasks',
+                            'service_config_settings', 'host_prep_tasks',
+                            'metadata_settings', 'kolla_config',
+                            'global_config_settings', 'logging_source',
+                            'logging_groups', 'external_deploy_tasks',
+                            'external_post_deploy_tasks',
+                            'docker_config_scripts', 'step_config']
 REQUIRED_DOCKER_PUPPET_CONFIG_SECTIONS = ['config_volume', 'step_config',
                                           'config_image']
 OPTIONAL_DOCKER_PUPPET_CONFIG_SECTIONS = [ 'puppet_tags', 'volumes' ]
+REQUIRED_DOCKER_LOGGING_OUTPUTS = ['config_settings', 'docker_config',
+                                   'volumes', 'host_prep_tasks']
 # Mapping of parameter names to a list of the fields we should _not_ enforce
 # consistency across files on.  This should only contain parameters whose
 # definition we cannot change for backwards compatibility reasons.  New
@@ -113,7 +127,8 @@ PARAMETER_DEFINITION_EXCLUSIONS = {'CephPools': ['description',
                                    'ControllerExtraConfig': ['description'],
                                    'NovaComputeExtraConfig': ['description'],
                                    'controllerExtraConfig': ['description'],
-                                   'DockerSwiftConfigImage': ['default']
+                                   'DockerSwiftConfigImage': ['default'],
+                                   'input_values': ['default']
                                    }
 
 PREFERRED_CAMEL_CASE = {
@@ -137,6 +152,17 @@ VALIDATE_DOCKER_OVERRIDE = {
   # docker/service/sshd.yaml is a variation of the puppet sshd service
   './docker/services/sshd.yaml': False,
 }
+DEPLOYMENT_RESOURCE_TYPES = [
+    'OS::Heat::SoftwareDeploymentGroup',
+    'OS::Heat::StructuredDeploymentGroup',
+    'OS::Heat::StructuredDeployments',
+    'OS::Heat::SoftwareDeployments',
+    'OS::Heat::SoftwareDeployment',
+    'OS::Heat::StructuredDeployment',
+    'OS::TripleO::SoftwareDeployment'
+]
+
+VALID_ANSIBLE_UPGRADE_TAGS = [ 'common', 'validation', 'pre-upgrade' ]
 
 def exit_usage():
     print('Usage %s <yaml file or directory>' % sys.argv[0])
@@ -199,11 +225,112 @@ def validate_hci_computehci_role(hci_role_filename, hci_role_tpl):
             hci_role_services = role['ServicesDefault']
             hci_role_services.remove('OS::TripleO::Services::CephOSD')
             if sorted(hci_role_services) != sorted(compute_role_services):
-                print('ERROR: ServicesDefault in %s is different from'
+                print('ERROR: ServicesDefault in %s is different from '
                       'ServicesDefault in roles/Compute.yaml' % hci_role_filename)
                 return 1
     return 0
 
+def validate_hci_role(hci_role_filename, hci_role_tpl):
+    role_files = ['HciCephAll', 'HciCephFile', 'HciCephMon', 'HciCephObject']
+    if hci_role_filename in ['./roles/'+ x +'.yaml' for x in role_files]:
+        compute_role_filename = os.path.join(os.path.dirname(hci_role_filename),
+                                             './Compute.yaml')
+        compute_role_tpl = yaml.load(open(compute_role_filename).read())
+        compute_role_services = compute_role_tpl[0]['ServicesDefault']
+        for role in hci_role_tpl:
+            if role['name'] == 'HciCephAll':
+                hci_role_services = role['ServicesDefault']
+                hci_role_services.remove('OS::TripleO::Services::CephMds')
+                hci_role_services.remove('OS::TripleO::Services::CephMgr')
+                hci_role_services.remove('OS::TripleO::Services::CephMon')
+                hci_role_services.remove('OS::TripleO::Services::CephRbdMirror')
+                hci_role_services.remove('OS::TripleO::Services::CephRgw')
+                hci_role_services.remove('OS::TripleO::Services::CephOSD')
+            if role['name'] == 'HciCephFile':
+                hci_role_services = role['ServicesDefault']
+                hci_role_services.remove('OS::TripleO::Services::CephMds')
+                hci_role_services.remove('OS::TripleO::Services::CephOSD')
+            if role['name'] == 'HciCephMon':
+                hci_role_services = role['ServicesDefault']
+                hci_role_services.remove('OS::TripleO::Services::CephMgr')
+                hci_role_services.remove('OS::TripleO::Services::CephMon')
+                hci_role_services.remove('OS::TripleO::Services::CephOSD')
+            if role['name'] == 'HciCephObject':
+                hci_role_services = role['ServicesDefault']
+                hci_role_services.remove('OS::TripleO::Services::CephRgw')
+                hci_role_services.remove('OS::TripleO::Services::CephOSD')
+            if sorted(hci_role_services) != sorted(compute_role_services):
+                print('ERROR: ServicesDefault in %s is different from '
+                      'ServicesDefault in roles/Compute.yaml' % hci_role_filename)
+                return 1
+    return 0
+
+def validate_ceph_role(ceph_role_filename, ceph_role_tpl):
+    role_files = ['CephAll', 'CephFile', 'CephMon', 'CephObject']
+    if ceph_role_filename in ['./roles/'+ x +'.yaml' for x in role_files]:
+        ceph_storage_role_filename = os.path.join(os.path.dirname(ceph_role_filename),
+                                             './CephStorage.yaml')
+        ceph_storage_role_tpl = yaml.load(open(ceph_storage_role_filename).read())
+        ceph_storage_role_services = ceph_storage_role_tpl[0]['ServicesDefault']
+        for role in ceph_role_tpl:
+            if role['name'] == 'CephAll':
+                ceph_role_services = role['ServicesDefault']
+                ceph_role_services.remove('OS::TripleO::Services::CephClient')
+                ceph_role_services.remove('OS::TripleO::Services::CephMds')
+                ceph_role_services.remove('OS::TripleO::Services::CephMgr')
+                ceph_role_services.remove('OS::TripleO::Services::CephMon')
+                ceph_role_services.remove('OS::TripleO::Services::CephRbdMirror')
+                ceph_role_services.remove('OS::TripleO::Services::CephRgw')
+            if role['name'] == 'CephFile':
+                ceph_role_services = role['ServicesDefault']
+                ceph_role_services.remove('OS::TripleO::Services::CephClient')
+                ceph_role_services.remove('OS::TripleO::Services::CephMds')
+            if role['name'] == 'CephObject':
+                ceph_role_services = role['ServicesDefault']
+                ceph_role_services.remove('OS::TripleO::Services::CephClient')
+                ceph_role_services.remove('OS::TripleO::Services::CephRgw')
+            if sorted(ceph_role_services) != sorted(ceph_storage_role_services):
+                print('ERROR: ServicesDefault in %s is different from '
+                      'ServicesDefault in roles/Ceph_storage.yaml' % ceph_role_filename)
+                return 1
+    return 0
+
+def validate_controller_no_ceph_role(filename, tpl):
+    control_role_filename = os.path.join(os.path.dirname(filename),
+                                         './Controller.yaml')
+    control_role_tpl = yaml.load(open(control_role_filename).read())
+    control_role_services = control_role_tpl[0]['ServicesDefault']
+    for role in tpl:
+        if role['name'] == 'ControllerNoCeph':
+            services = role['ServicesDefault']
+            services.remove('OS::TripleO::Services::CephClient')
+            services.append('OS::TripleO::Services::CephMds')
+            services.append('OS::TripleO::Services::CephMgr')
+            services.append('OS::TripleO::Services::CephMon')
+            services.append('OS::TripleO::Services::CephRbdMirror')
+            services.append('OS::TripleO::Services::CephRgw')
+            if sorted(services) != sorted(control_role_services):
+                print('ERROR: ServicesDefault in %s is different from '
+                      'ServicesDefault in roles/Controller.yaml' % filename)
+                return 1
+    return 0
+
+def validate_with_compute_role_services(role_filename, role_tpl, exclude_service):
+    cmpt_filename = os.path.join(os.path.dirname(role_filename),
+                                 './Compute.yaml')
+    cmpt_tpl = yaml.load(open(cmpt_filename).read())
+    cmpt_services = cmpt_tpl[0]['ServicesDefault']
+    cmpt_services = [x for x in cmpt_services if (x not in exclude_service)]
+
+    role_services = set(role_tpl[0]['ServicesDefault'])
+    missing_services = list(set(cmpt_services) - role_services)
+    if missing_services:
+        print('ERROR: ServicesDefault in {0} is missing services [{1}] from '
+              'ServicesDefault in roles/Compute.yaml'.format(role_filename,
+              ', '.join(missing_services)))
+        return 1
+    return 0
+
 
 def search(item, check_item, check_key):
     if check_item(item):
@@ -315,6 +442,12 @@ def validate_docker_service(filename, tpl):
 
         for section_name in REQUIRED_DOCKER_SECTIONS:
             if section_name not in role_data:
+                # add an exception if both step_config is used in docker
+                # service, docker/services/ceph-ansible/ceph-nfs.yaml uses
+                # additional step_config to add pacemaker resources
+                if (section_name == 'docker_config' and
+                        role_data.get('step_config', '')):
+                    continue
                 print('ERROR: %s is required in role_data for %s.'
                       % (section_name, filename))
                 return 1
@@ -380,6 +513,17 @@ def validate_docker_service(filename, tpl):
                       print('ERROR: bootstrap_host_exec needs to run as the root user.')
                       return 1
 
+        if 'upgrade_tasks' in role_data and role_data['upgrade_tasks']:
+            if (validate_upgrade_tasks(role_data['upgrade_tasks']) or
+                validate_upgrade_tasks_duplicate_whens(filename)):
+                print('ERROR: upgrade_tasks validation failed')
+                return 1
+
+        if 'fast_forward_upgrade_tasks' in role_data and role_data['fast_forward_upgrade_tasks']:
+            if validate_upgrade_tasks(role_data['fast_forward_upgrade_tasks']):
+                print('ERROR: fast_forward_upgrade_tasks validation failed')
+                return 1
+
     if 'parameters' in tpl:
         for param in required_params:
             if param not in tpl['parameters']:
@@ -389,6 +533,20 @@ def validate_docker_service(filename, tpl):
     return 0
 
 
+def validate_docker_logging_template(filename, tpl):
+    if 'outputs' not in tpl:
+        print('ERROR: outputs are missing from: %s' % filename)
+        return 1
+    missing_entries = [
+        entry for entry in REQUIRED_DOCKER_LOGGING_OUTPUTS
+        if entry not in tpl['outputs']]
+    if any(missing_entries):
+        print('ERROR: The file %s is missing the following output(s):'
+              ' %s' % (filename, ', '.join(missing_entries)))
+        return 1
+    return 0
+
+
 def validate_service(filename, tpl):
     if 'outputs' in tpl and 'role_data' in tpl['outputs']:
         if 'value' not in tpl['outputs']['role_data']:
@@ -412,6 +570,17 @@ def validate_service(filename, tpl):
            validate_mysql_connection(role_data['config_settings']):
             print('ERROR: mysql connection uri should use option bind_address')
             return 1
+        if 'upgrade_tasks' in role_data and role_data['upgrade_tasks']:
+            if (validate_upgrade_tasks(role_data['upgrade_tasks']) or
+                validate_upgrade_tasks_duplicate_whens(filename)):
+                print('ERROR: upgrade_tasks validation failed')
+                return 1
+
+        if 'fast_forward_upgrade_tasks' in role_data and role_data['fast_forward_upgrade_tasks']:
+            if validate_upgrade_tasks(role_data['fast_forward_upgrade_tasks']):
+                print('ERROR: fast_forward_upgrade_tasks validation failed')
+                return 1
+
     if 'parameters' in tpl:
         for param in required_params:
             if param not in tpl['parameters']:
@@ -421,6 +590,31 @@ def validate_service(filename, tpl):
     return 0
 
 
+
+def validate_upgrade_tasks_duplicate_whens(filename):
+    """Take a heat template and starting at the upgrade_tasks
+       try to detect duplicate 'when:' statements
+    """
+    with open(filename) as template:
+        contents = template.read()
+        upgrade_task_position = contents.index('upgrade_tasks')
+        lines = contents[upgrade_task_position:].splitlines()
+        count = 0
+        duplicate = ''
+        for line in lines:
+            if '  when:' in line:
+                count += 1
+                if count > 1:
+                    print ("ERROR: found duplicate when statements in %s "
+                           "upgrade_task: %s %s" % (filename, line, duplicate))
+                    return 1
+                duplicate = line
+            elif ' -' in line:
+                count = 0
+                duplicate = ''
+        return 0
+
+
 def validate(filename, param_map):
     """Validate a Heat template
 
@@ -445,24 +639,47 @@ def validate(filename, param_map):
                                 ...
                            ]}
     """
-    print('Validating %s' % filename)
+    if args.quiet < 1:
+        print('Validating %s' % filename)
     retval = 0
     try:
         tpl = yaml.load(open(filename).read())
 
-        # The template alias version should be used instead a date, this validation
-        # will be applied to all templates not just for those in the services folder.
-        if 'heat_template_version' in tpl and not str(tpl['heat_template_version']).isalpha():
-            print('ERROR: heat_template_version needs to be the release alias not a date: %s'
-                  % filename)
-            return 1
+        is_heat_template = 'heat_template_version' in tpl
+
+        # The template version should be in the list of supported versions for the current release.
+        # This validation will be applied to all templates not just for those in the services folder.
+        if is_heat_template:
+            tpl_template_version = str(tpl['heat_template_version'])
+            if tpl_template_version not in valid_heat_template_versions:
+                print('ERROR: heat_template_version in template %s '
+                      'is not valid: %s (allowed values %s)'
+                    % (
+                        filename,
+                        tpl_template_version,
+                        ', '.join(valid_heat_template_versions)
+                    )
+                )
+                return 1
+            if tpl_template_version != current_heat_template_version:
+                print('Warning: heat_template_version in template %s '
+                      'is outdated: %s (current %s)'
+                    % (
+                        filename,
+                        tpl_template_version,
+                        current_heat_template_version
+                    )
+                )
 
         if VALIDATE_PUPPET_OVERRIDE.get(filename, False) or (
                 filename.startswith('./puppet/services/') and
                 VALIDATE_PUPPET_OVERRIDE.get(filename, True)):
             retval = validate_service(filename, tpl)
 
-        if VALIDATE_DOCKER_OVERRIDE.get(filename, False) or (
+
+        if filename.startswith('./docker/services/logging/'):
+            retval = validate_docker_logging_template(filename, tpl)
+        elif VALIDATE_DOCKER_OVERRIDE.get(filename, False) or (
                 filename.startswith('./docker/services/') and
                 VALIDATE_DOCKER_OVERRIDE.get(filename, True)):
             retval = validate_docker_service(filename, tpl)
@@ -473,11 +690,42 @@ def validate(filename, param_map):
         if filename.startswith('./roles/ComputeHCI.yaml'):
             retval = validate_hci_computehci_role(filename, tpl)
 
+        if filename.startswith('./roles/ComputeOvsDpdk.yaml') or \
+                filename.startswith('./roles/ComputeSriov.yaml') or \
+                filename.startswith('./roles/ComputeOvsDpdkRT.yaml') or \
+                filename.startswith('./roles/ComputeSriovRT.yaml'):
+            exclude = [
+                'OS::TripleO::Services::OVNController',
+                'OS::TripleO::Services::ComputeNeutronOvsAgent',
+                'OS::TripleO::Services::Tuned',
+                'OS::TripleO::Services::NeutronVppAgent',
+                'OS::TripleO::Services::Vpp',
+                'OS::TripleO::Services::NeutronLinuxbridgeAgent']
+            retval = validate_with_compute_role_services(filename, tpl, exclude)
+
+        if filename.startswith('./roles/ComputeRealTime.yaml'):
+            exclude = [
+                'OS::TripleO::Services::Tuned',
+            ]
+            retval = validate_with_compute_role_services(filename, tpl, exclude)
+
+        if filename.startswith('./roles/Hci'):
+            retval = validate_hci_role(filename, tpl)
+
+        if filename.startswith('./roles/Ceph'):
+            retval = validate_ceph_role(filename, tpl)
+
+        if filename.startswith('./roles/ControllerNoCeph.yaml'):
+            retval = validate_controller_no_ceph_role(filename, tpl)
+
+        if filename.startswith('./network_data_'):
+            retval = validate_network_data_file(filename)
+
     except Exception:
         print(traceback.format_exc())
         return 1
     # yaml is OK, now walk the parameters and output a warning for unused ones
-    if 'heat_template_version' in tpl:
+    if is_heat_template:
         for p, data in tpl.get('parameters', {}).items():
             definition = {'data': data, 'filename': filename}
             param_map.setdefault(p, []).append(definition)
@@ -486,16 +734,80 @@ def validate(filename, param_map):
             str_p = '\'%s\'' % p
             in_resources = str_p in str(tpl.get('resources', {}))
             in_outputs = str_p in str(tpl.get('outputs', {}))
-            if not in_resources and not in_outputs:
+            if not in_resources and not in_outputs and args.quiet < 2:
                 print('Warning: parameter %s in template %s '
                       'appears to be unused' % (p, filename))
 
+        resources = tpl.get('resources')
+        if resources:
+            for resource, data in resources.items():
+                if data['type'] not in DEPLOYMENT_RESOURCE_TYPES:
+                    continue
+                if 'name' not in data['properties']:
+                    print('ERROR: resource %s from %s missing name property.'
+                            % (resource, filename))
+                    return 1
+
     return retval
 
-if len(sys.argv) < 2:
-    exit_usage()
+def validate_upgrade_tasks(upgrade_tasks):
+    # some templates define its upgrade_tasks via list_concat
+    if isinstance(upgrade_tasks, dict):
+        if upgrade_tasks.get('list_concat'):
+            return validate_upgrade_tasks(upgrade_tasks['list_concat'][1])
+        elif upgrade_tasks.get('get_attr'):
+            return 0
 
-path_args = sys.argv[1:]
+    for task in upgrade_tasks:
+        task_name = task.get("name", "")
+        if task.get("tags"):
+            if (task["tags"] not in VALID_ANSIBLE_UPGRADE_TAGS):
+                print('ERROR: Task (%s) includes unexpected \'tags: (%s)\' ' % (task_name, task["tags"]))
+                return 1
+        else:
+
+            whenline = task.get("when", "")
+            if (type(whenline) == list):
+                if any('step|int ' in condition for condition in whenline) and ('step|int == ' not in whenline[0]):
+                        print('ERROR: \'step|int ==\' condition should be evaluated first in when conditions for task (%s)'  % (task))
+                        return 1
+            else:
+                if (' and ' in whenline) and (' or ' not in whenline):
+                    print("Warning: Consider specifying \'and\' conditions as a list to improve readability in task: \"%s\"" %  (task_name))
+    return 0
+
+def validate_network_data_file(data_file_path):
+    try:
+        data_file = yaml.load(open(data_file_path).read())
+        base_file_path = os.path.dirname(data_file_path) + "/network_data.yaml"
+        base_file = yaml.load(open(base_file_path).read())
+        for n in base_file:
+            if n not in data_file:
+                print('ERROR: The following network from network_data.yaml is '
+                      'missing or differs in %s : %s'
+                      % (data_file_path, n))
+                return 1
+    except Exception:
+        print(traceback.format_exc())
+        return 1
+    return 0
+
+def parse_args():
+    p = argparse.ArgumentParser()
+
+    p.add_argument('--quiet', '-q',
+                   action='count',
+                   default=0,
+                   help='output warnings and errors (-q) or only errors (-qq)')
+    p.add_argument('path_args',
+                   nargs='*',
+                   default=['.'])
+
+    return p.parse_args()
+
+args = parse_args()
+path_args = args.path_args
+quiet = args.quiet
 exit_val = 0
 failed_files = []
 base_endpoint_map = None
@@ -539,7 +851,7 @@ if base_endpoint_map and \
                   "endpoint map" % env_endpoint_map['file'])
             failed_files.append(env_endpoint_map['file'])
             exit_val |= 1
-        else:
+        elif args.quiet < 1:
             print("%s matches base endpoint map" % env_endpoint_map['file'])
 else:
     print("ERROR: Did not find expected number of environments containing the "
